import typing

import jax
import pydot

from .dot import draw_dot_graph


def draw(f, collapse_primitives=True, show_avals=True) -> typing.Callable:
    """
    Visualise a JAX computation graph

    Wraps a JAX jit compiled function, which when called
    visualises the computation graph using
    pydot.

    Examples
    --------

    .. highlight:: python
    .. code-block:: python

       import jax
       import jpviz

       @jax.jit
       def foo(x):
           return 2 * x

       @jax.jit
       def bar(x):
           x = foo(x)
           return x - 1

       g = jpviz.draw(bar)(jax.numpy.arange(10))

    Parameters
    ----------
    f:
        JAX jit compiled function
    collapse_primitives: bool
        If `True` sub-functions that contain only JAX primitives
        will be collapsed into a single node in the generated
        graph
    show_avals: bool
        If `True` then type information will be
        included on node labels

    Returns
    -------
    Wrapped function that when called with concrete
    values generated the corresponding visualisation
    of the computation graph
    """

    def _inner_draw(*args, **kwargs) -> pydot.Graph:
        jaxpr = jax.make_jaxpr(f)(*args, **kwargs)
        return draw_dot_graph(jaxpr, collapse_primitives, show_avals)

    return _inner_draw


def view_pydot(dot_graph: pydot.Dot) -> None:
    """
    Show a pydot graph in a jupyter notebook

    Parameters
    ----------
    dot_graph: Graph
        Pydot graph as generated by `draw`
    """
    from IPython.display import Image, display

    plt = Image(dot_graph.create_png())
    display(plt)



import pydot
from jax._src import core as jax_core

from . import graph


def draw_dot_graph(
    fn: jax_core.ClosedJaxpr, collapse_primitives: bool, show_avals: bool
) -> pydot.Graph:
    """
    Generate a pydot representation of an XLA graph

    Parameters
    ----------
    fn : ClosedJaxpr
    collapse_primitives: bool
        If `True` functions that are composed of only primitive
        values will be collapsed
    show_avals: bool
        If `True` type information will be included in the node label

    Returns
    -------
    Graph
        Pydot graph
    """

    g = pydot.Dot(graph_type="digraph")

    sub_graph, _, _, _, _ = graph.get_sub_graph(
        fn.eqns[0], "", 0, collapse_primitives, show_avals
    )
    if isinstance(sub_graph, pydot.Subgraph):
        g.add_subgraph(sub_graph)
    else:
        g.add_node(sub_graph)

    return g



import typing

from jax._src import core as jax_core


def get_node_label(
    v: typing.Union[jax_core.Var, jax_core.Literal], show_avals: bool
) -> str:
    """
    Concatenate a variable name and its type.

    Parameters
    ----------
    v: Var
        Jax variable
    show_avals: bool
        If `True` then the type will be included in the
        node label

    Returns
    -------
    str
    """
    if show_avals:
        return f"{v}: {v.aval.str_short()}"
    else:
        return str(v)


def is_not_primitive(x: jax_core.JaxprEqn) -> bool:
    """
    Test if a JaxprEqn is a primitive.

    Parameters
    ----------
    x: JaxprEqn

    Returns
    -------
    bool
        'True' if not a primitive.
    """
    return "jaxpr" in x.params


def contains_non_primitives(eqns: typing.List[jax_core.JaxprEqn]) -> bool:
    """
    Check it the sub-functions of a JaxPR contains only JAX primitives

    Parameters
    ----------
    eqns: List[jax._src.core.JaxprEqn]
        List of JaxprEqns

    Returns
    -------
    bool:
        `True` if any of the sub-eqns are non-primitive
    """
    return any([("jaxpr" in e.params or e.primitive.name == "cond") for e in eqns])
